import os
from types import FunctionType

PYODIDE_VERSION = "0.27.7"

def route_to_view(code: str, route: str) -> FunctionType:
	frontend_header = """import micropip
await micropip.install('volumetric-flask')
import pyjsx.auto_setup
"""

	code_wrapped = f"exec({code.encode()!r}.decode('jsx'), globals(), locals())\n"

	frontend_footer = "import volumetric\nupdate" # return value to JS

	frontend_code = frontend_header+code_wrapped+frontend_footer

	py_residence_dir = f"static/_python{route}"

	if not os.path.exists(py_residence_dir):
		os.makedirs(py_residence_dir)

	py_residence = os.path.join(py_residence_dir, "view.py")

	open(py_residence, 'w').write(frontend_code)

	# TODO: optimize DOM updates

	frontend_html = f"""<!DOCTYPE html>
<html>
	<head>
		<script src="https://cdn.jsdelivr.net/pyodide/v{PYODIDE_VERSION}/full/pyodide.js"></script>
		<script>
			const pyodidePromise = loadPyodide();

			async function main() {{
				const pyodide = await pyodidePromise;
				
				await pyodide.loadPackage("micropip")

				const resource = await fetch("{py_residence}");
				const code = await resource.text();
				const update = await pyodide.runPythonAsync(code);

				console.log("...");

				const volumetric = pyodide.pyimport("volumetric");

				console.log("done!");
				
				volumetric.csr.clear_convd();
				document.documentElement.innerHTML = update();
				console.log("done2");
				
				// while (true) {{
				//	
				// }}
			}}

			async function callconvd(name) {{
				const pyodide = await loadPyodide();
			
				const volumetric = pyodide.pyimport("volumetric");
				
				volumetric.csr.convd_functions[name]();
			}}

			main();
		</script>
	</head>
	<body></body>
</html>"""
	
	return lambda: frontend_html

convd_functions = {}

def clear_convd():
	convd_functions.clear()

def convert_function(func: FunctionType) -> str:
	convd_functions[repr(func)] = func

	return f"callconvd({repr(func)!r})"


import os
import shutil
import sys
from flask import Flask
from json import load, dumps
from shutil import rmtree
from sys import argv
from importlib import import_module
from argparse import ArgumentParser

WATCH_FILES: dict[str, int] = {}

def run(force_debug: bool):
	conf = {}

	if os.path.exists("config.json"):
		with open("config.json", 'r') as f:
			conf: dict = load(f)

	if not os.path.isfile(f"app.py"):
		print("app.py file does not exist!")
		exit(1)

	sys.path.insert(0, os.getcwd()) # this is needed in case the script was run without `python -m`

	appmod = import_module("app")
	
	try:
		app: Flask = appmod.app
	except AttributeError:
		print("app object is missing from app.py!")
		exit(1)

	if force_debug: app.debug = True
		
	app.run(**conf)

DEFAULT_SSR_ROUTE_CODE = """# coding: jsx

from pyjsx import jsx
import volumetric
from volumetric.xml_helpers import body

def handler(app: volumetric.App, *args):	
	return body(
		<h1 id="heading"></h1>,
		head=<>
			<script src="/static/js/index.js" defer></script>
			<link rel="stylesheet" href="/static/css/index.css"/>
		</>
	)
"""

DEFAULT_CSR_ROUTE_CODE = """from js import *
from pyjsx import jsx
from volumetric.xml_helpers import body

def update():
	return body(
		<h1 id="heading"></h1>,
		head=<>
			<script src="/static/js/index.js" defer></script>
			<link rel="stylesheet" href="/static/css/index.css"/>
		</>
	)

"""

DEFAULT_SSR_ROUTE_CONF = {
	"methods": ["GET"]
}

DEFAULT_CSR_ROUTE_CONF = {
	"methods": ["GET"],
	"csr": True
}

DEFAULT_CONF = {
	"host": "127.0.0.1",
	"port": 5000
}

DEFAULT_CODE = """from volumetric import App

app = App(__name__)

app.fs_routes.enable()

app.debug = True"""

def new(name: str):
	if os.path.exists(name): rmtree(name)

	os.mkdir(name)

	os.mkdir(f"{name}/static")

	os.mkdir(f"{name}/static/js")
	open(f"{name}/static/js/index.js", 'w').write(
		"""document.getElementById("heading")
	.textContent = "Hello World!"
"""
	)

	os.mkdir(f"{name}/static/css")
	open(f"{name}/static/css/index.css", 'w').write(
		"""#heading {
	color: red;
}"""
	)

	os.mkdir(f"{name}/static/images")

	open(f"{name}/app.py", 'w').write(
		DEFAULT_CODE
	)

	open(f"{name}/config.json", 'w').write(
		dumps(DEFAULT_CONF, indent='\t')
	)

	os.mkdir(f"{name}/root")
	
	open(f"{name}/root/config.json", 'w').write(
		dumps(DEFAULT_SSR_ROUTE_CONF, indent='\t')
	)

	open(f"{name}/root/index.py", 'w').write(
		DEFAULT_SSR_ROUTE_CODE
	)

def route(path: str, csr: bool):
	if not os.path.isdir("root"):
		print("volumetric route must be called from the app directory!")
		exit(1)

	if not path.startswith('/'):
		print("pathname must start from root path! (/)")
		exit(1)

	project_path = f"root{path}"
	
	if os.path.exists(project_path): rmtree(project_path)

	os.makedirs(project_path)
	
	if csr:
		open(f"{project_path}/config.json", 'w').write(
			dumps(DEFAULT_CSR_ROUTE_CONF, indent='\t')
		)

		open(f"{project_path}/index.py", 'w').write(
			DEFAULT_CSR_ROUTE_CODE
		)	
	else:
		open(f"{project_path}/config.json", 'w').write(
			dumps(DEFAULT_SSR_ROUTE_CONF, indent='\t')
		)

		open(f"{project_path}/index.py", 'w').write(
			DEFAULT_SSR_ROUTE_CODE
		)

def cleancsr():
	shutil.rmtree("static/_python")
	print("Removed static/_python/")

def main():
	parser = ArgumentParser("volumetric", description="CLI for the Volumetric Python web framework (docs: https://DOCS_SUBDOMAIN.readthedocs.io/)")
	
	subparsers = parser.add_subparsers(dest="command", required=True, help="sub-command help")

	new_parser = subparsers.add_parser(
		"new",
		help="Create a new project"
	)
	
	new_parser.add_argument(
		"name",
		help="Name to be used for the new project"
	)

	route_parser = subparsers.add_parser(
		"route",
		help="Create a new route directory"
	)

	route_parser.add_argument(
		"path",
		help="Name to be used for the new route"
	)

	route_parser.add_argument(
		"--csr",
		help="Make this route client-side rendered",
		action="store_true"
	)

	run_parser = subparsers.add_parser(
		"run",
		help="Run the project"
	)

	run_parser.add_argument("--force-debug", action="store_true", help="make sure debug mode is used")

	clearcsr_parser = subparsers.add_parser(
		"cleancsr",
		help="Clean Python files generated at startup for CSR"
	)

	args = parser.parse_args()

	if args.command == "route":		
		route(args.path, args.csr)
	elif args.command == "new":
		new(args.name)
	elif args.command == "cleancsr":
		cleancsr()
	elif args.command == "run":
		run(args.force_debug)

if __name__ == "__main__":
	main()


# coding: jsx

from typing import Callable
from pyjsx import jsx, JSX

def Content(children: list[JSX], **props):
	props["class"] = ("content "+props.get("class", "")).strip()
	
	return <div {...props}>{children}</div>


import pyjsx.auto_setup
import os
import typing as t
import dotenv
from flask import Flask

from volumetric.fs_routes import parse_fs_routes
	
class SecretsProxy:
	def __getattr__(self, attr: str):
		return os.environ[attr]
	
class PluginObjects: pass

class FSRoutesManager:
	def __init__(self, app: 'App'):
		self.app = app

	def enable(self):
		if not parse_fs_routes(self.app, "root"):
			exit(1)

class App(Flask):
	def __init__(
		self,
		import_name: str,
		static_url_path: t.Optional[str] = None,
		static_folder: t.Optional[t.Union[str, os.PathLike]] = "static",
		static_host: t.Optional[str] = None,
		host_matching: bool = False,
		subdomain_matching: bool = False,
		template_folder: t.Optional[str] = "templates",
		instance_path: t.Optional[str] = None,
		instance_relative_config: bool = False,
		root_path: t.Optional[str] = None,
		secrets_path: t.Optional[str] = None
	):
		super().__init__(
			import_name, 
			static_url_path, 
			static_folder, 
			static_host, 
			host_matching, 
			subdomain_matching, 
			template_folder,
			instance_path,
			instance_relative_config,
			root_path
		)

		if secrets_path:
			dotenv.load_dotenv(f"{self.instance_path}/{secrets_path}")
			
			self.secrets = SecretsProxy()

		self.plugin_objects = PluginObjects()
		self.fs_routes = FSRoutesManager(self)


from functools import partial
from importlib import import_module
import os
import json
import sys
from types import FunctionType
from flask import Flask

from volumetric import csr

def appbind(func: FunctionType, app: Flask, name: str):
	new = partial(func, app)
	new.__name__ = name

	return new

def modularize_index(path: str):
	dirname = os.path.dirname(path)

	sys.path.insert(0, dirname)

	mod = import_module("index")

	sys.path.remove(dirname)

	return mod
	

def parse_fs_routes(app: Flask, rootdir: str, parent: str='/') -> bool:
	conf_fname = f"{rootdir}/config.json"
	index_fname = f"{rootdir}/index.py"
	
	if os.path.exists(conf_fname):
		with open(conf_fname, 'r') as f:
			try: config: dict = json.load(f)
			except json.decoder.JSONDecodeError:
				print(f"{conf_fname} is invalid!")
				return False
	else: config = {}

	if os.path.exists(index_fname):
		if config.get("csr") is True:
			handler = csr.route_to_view(open(index_fname).read(), parent)
			del config["csr"]
		else:
			try:
				index = modularize_index(index_fname)
			except Exception as e:
				print(f"{index_fname} threw an error!\n{e}")
				return False

			if not hasattr(index, "handler"):
				print(f"{index_fname} is missing a handler function!")
				return False
			
			handler = appbind(
				index.handler,
				app,
				f"{parent}_handler"
			)

			del sys.modules["index"]

		app.route(parent, **config)(handler)

	for subdir in os.listdir(rootdir):
		sub_qual = f"{rootdir}/{subdir}"
		if os.path.isdir(sub_qual):
			if not parse_fs_routes(app, sub_qual, f"{parent}{subdir}/"):
				return False

	return True




# coding: jsx

from typing import Any, Callable

from pyjsx import jsx, JSX


def IfTruthy(children: list[JSX], valid: Any | None, action: Callable[[], None]=lambda: None) -> JSX:
	if valid:
		action()
		return children

	return ""

def NotTruthy(children: list[JSX], valid: Any | None, action: Callable[[], None]=lambda: None) -> JSX:
	if not valid:
		action()
		return children

	return ""



import html
from typing import Any
from pyjsx import JSX

def body(elem: JSX, head: JSX=""):
	return f"""<html><head>{head}</head><body>{elem}</body></html>"""

def escape_repr(obj: Any):
	return html.escape(repr(obj), quote=True)


# coding: jsx

from typing import Any, Callable

from pyjsx import jsx, JSX


def Try(children: list[JSX], func: Callable[[], JSX], catch: Exception | tuple[Exception], fallback: JSX) -> JSX:
	try:
		return func()
	except catch:
		return fallback
